# -*- coding: utf-8 -*-
import atexit
import os
import shutil
import tempfile
from typing import Any, Dict, Union

from kiara import KiaraModule, KiaraModuleConfig, ValueMap, ValueMapSchema
from kiara.models.filesystem import FileBundle, FileModel
from pydantic import Field


class ImportFileConfig(KiaraModuleConfig):

    import_metadata: bool = Field(
        description="Whether to return the import metadata as well.",
        default=True,
    )


class ImportFileModule(KiaraModule):
    """A generic module to import a file from any local or remote location."""

    _module_type_name = "import.file"
    _config_cls = ImportFileConfig

    def create_inputs_schema(
        self,
    ) -> ValueMapSchema:

        result: Dict[str, Dict[str, Any]] = {
            "uri": {
                "type": "string",
                "doc": "The uri (url/path/...) of the file to import.",
            }
        }
        if self.get_config_value("import_metadata"):
            result["import_metadata"] = {
                "type": "dict",
                "doc": "Metadata you want to attach to the file.",
                "optional": True,
            }

        return result

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        result = {
            "file": {
                "type": "file",
                "doc": "The imported file.",
            }
        }
        if self.get_config_value("import_metadata"):
            result["import_metadata"] = {
                "type": "dict",
                "doc": "Metadata about the import and file.",
            }
        return result

    def process(self, inputs: ValueMap, outputs: ValueMap) -> None:
        pass


class ImportFileBundleConfig(KiaraModuleConfig):

    import_metadata: bool = Field(
        description="Whether to return the import metadata as well.",
        default=True,
    )


class ImportFileBundleModule(KiaraModule):
    """A generic module to import a file bundle from any local or remote location."""

    _module_type_name = "import.file_bundle"
    _config_cls = ImportFileBundleConfig

    def create_inputs_schema(
        self,
    ) -> ValueMapSchema:

        result: Dict[str, Dict[str, Any]] = {
            "uri": {
                "type": "string",
                "doc": "The uri (url/path/...) of the file to import.",
            }
        }
        if self.get_config_value("import_metadata"):
            result["import_metadata"] = {
                "type": "dict",
                "doc": "Metadata you want to attach to the file bundle.",
                "optional": True,
            }

        return result

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        result = {
            "file_bundle": {
                "type": "file_bundle",
                "doc": "The imported file bundle.",
            }
        }
        if self.get_config_value("import_metadata"):
            result["import_metadata"] = {
                "type": "dict",
                "doc": "Metadata about the import and file bundle.",
            }
        return result

    def process(self, inputs: ValueMap, outputs: ValueMap) -> None:
        pass


class DownloadFileConfig(KiaraModuleConfig):
    download_metadata: bool = Field(
        description="Whether to return the download metadata as well.",
        default=True,
    )


class DownloadFileModule(KiaraModule):
    """Download a single file from a remote location.

    The result of this operation is a single value of type 'file' (basically an array of raw bytes), which can then be used in other modules to
    create more meaningful data structures.
    """

    _module_type_name = "download.file"
    _config_cls = DownloadFileConfig

    def create_inputs_schema(self) -> ValueMapSchema:

        result: Dict[str, Dict[str, Any]] = {
            "url": {"type": "string", "doc": "The url of the file to download."},
            "file_name": {
                "type": "string",
                "doc": "The file name to use for the downloaded file.",
                "optional": True,
            },
        }
        return result

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        result: Dict[str, Dict[str, Any]] = {
            "file": {
                "type": "file",
                "doc": "The downloaded file.",
            }
        }

        if self.get_config_value("download_metadata"):
            result["download_metadata"] = {
                "type": "dict",
                "doc": "Metadata about the download.",
            }
        return result

    def process(self, inputs: ValueMap, outputs: ValueMap):

        from datetime import datetime

        import httpx
        import pytz

        url = inputs.get_value_data("url")
        file_name = inputs.get_value_data("file_name")
        tmp_file = tempfile.NamedTemporaryFile(delete=False)
        atexit.register(tmp_file.close)

        history = []
        datetime.utcnow().replace(tzinfo=pytz.utc)
        with open(tmp_file.name, "wb") as f:
            with httpx.stream("GET", url, follow_redirects=True) as r:
                history.append(dict(r.headers))
                for h in r.history:
                    history.append(dict(h.headers))
                for data in r.iter_bytes():
                    f.write(data)

        if not file_name:
            # TODO: make this smarter, using content-disposition headers if available
            file_name = url.split("/")[-1]

        result_file = FileModel.load_file(tmp_file.name, file_name)

        metadata = {
            "response_headers": history,
            "request_time": datetime.utcnow().replace(tzinfo=pytz.utc).isoformat(),
        }
        outputs.set_value("download_metadata", metadata)
        outputs.set_value("file", result_file)


class DownloadFileBundleModule(KiaraModule):
    _module_type_name = "download.file_bundle"
    _config_cls = DownloadFileConfig

    def create_inputs_schema(self) -> ValueMapSchema:

        result: Dict[str, Dict[str, Any]] = {
            "url": {
                "type": "string",
                "doc": "The url of an archive/zip file to download.",
            },
            "sub_path": {
                "type": "string",
                "doc": "A relative path to select only a sub-folder from the archive.",
                "optional": True,
            },
        }

        return result

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        result: Dict[str, Dict[str, Any]] = {
            "file_bundle": {
                "type": "file_bundle",
                "doc": "The downloaded file bundle.",
            }
        }

        if self.get_config_value("download_metadata"):
            result["download_metadata"] = {
                "type": "dict",
                "doc": "Metadata about the download.",
            }
        return result

    def process(self, inputs: ValueMap, outputs: ValueMap):

        from datetime import datetime

        import httpx
        import pytz

        url = inputs.get_value_data("url")
        sub_path: Union[None, str] = inputs.get_value_data("sub_path")
        tmp_file = tempfile.NamedTemporaryFile(delete=False)
        atexit.register(tmp_file.close)

        history = []
        datetime.utcnow().replace(tzinfo=pytz.utc)
        with open(tmp_file.name, "wb") as f:
            with httpx.stream("GET", url, follow_redirects=True) as r:
                history.append(dict(r.headers))
                for h in r.history:
                    history.append(dict(h.headers))
                for data in r.iter_bytes():
                    f.write(data)

        import patoolib

        out_dir = tempfile.mkdtemp()

        def del_out_dir():
            shutil.rmtree(out_dir, ignore_errors=True)

        # atexit.register(del_out_dir)

        patoolib.extract_archive(tmp_file.name, outdir=out_dir)

        path = out_dir
        if sub_path:
            path = os.path.join(out_dir, sub_path)
        bundle = FileBundle.import_folder(path)

        metadata = {
            "response_headers": history,
            "request_time": datetime.utcnow().replace(tzinfo=pytz.utc).isoformat(),
        }
        outputs.set_value("download_metadata", metadata)
        outputs.set_value("file_bundle", bundle)



# -*- coding: utf-8 -*-

"""This module contains the metadata (and other) models that are used in the ``kiara_plugin.onboarding`` package.

Those models are convenience wrappers that make it easier for *kiara* to find, create, manage and version metadata -- but also
other type of models -- that is attached to data, as well as *kiara* modules.

Metadata models must be a sub-class of [kiara.metadata.MetadataModel][kiara.metadata.MetadataModel]. Other models usually
sub-class a pydantic BaseModel or implement custom base classes.
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys

from setuptools import setup

try:
    from pkg_resources import VersionConflict, require

    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


def get_extra_requires(add_all=True, add_all_dev=True):

    from distutils.dist import Distribution

    dist = Distribution()
    dist.parse_config_files()
    dist.parse_command_line()

    extras = {}
    extra_deps = dist.get_option_dict("options.extras_require")

    for extra_name, data in extra_deps.items():

        _, dep_string = data
        deps = []
        d = dep_string.split("\n")
        for line in d:
            if not line:
                continue
            deps.append(line)
        extras[extra_name] = deps

    if add_all:
        all = set()
        for e_n, deps in extras.items():
            if not e_n.startswith("dev_"):
                all.update(deps)
        extras["all"] = all

    # add tag `all` at the end
    if add_all_dev:
        extras["all_dev"] = set(vv for v in extras.values() for vv in v)
        extras["dev_all"] = extras["all_dev"]

    return extras


if __name__ in ["__main__", "builtins", "__builtin__"]:
    setup(
        use_scm_version={"write_to": "src/kiara_plugin/onboarding/version.txt"},
        extras_require=get_extra_requires(),
    )



# -*- coding: utf-8 -*-
import hashlib
import shutil
from pathlib import Path
from typing import Any, Mapping

import orjson
from kiara import KiaraModule, KiaraModuleConfig, ValueMap, ValueMapSchema
from kiara.exceptions import KiaraProcessingException
from kiara.models.filesystem import FileBundle
from pydantic import Field


class ZenodoDownloadConfig(KiaraModuleConfig):

    metadata_filename: str = Field(
        description="The filename for the zenodo metadata.", default="metadata.json"
    )


class ZenodoDownload(KiaraModule):
    """Download a dataset from zenodo.org."""

    _module_type_name = "onboard.zenodo_record"
    _config_cls = ZenodoDownloadConfig

    def create_inputs_schema(
        self,
    ) -> ValueMapSchema:

        metadata_filename = self.get_config_value("metadata_filename")
        return {
            "doi": {"type": "string", "doc": "The doi of the record"},
            "include_metadata": {
                "type": "boolean",
                "doc": f"Whether to write the record metadata to a file '{metadata_filename}' and include it in the resulting file bundle.",
                "default": True,
            },
        }

    def create_outputs_schema(
        self,
    ) -> ValueMapSchema:

        return {
            "file_bundle": {
                "type": "file_bundle",
            }
        }

    def download_file(self, file_data: Mapping[str, Any], target_path: Path):

        import httpx

        url = file_data["links"]["self"]
        file_name = file_data["key"]
        checksum = file_data["checksum"][4:]

        target_file = target_path / file_name

        if target_file.exists():
            raise KiaraProcessingException(
                f"Can't download file, target path already exists: {target_path.as_posix()}."
            )

        hash_md5 = hashlib.md5()

        with open(target_file, "ab") as file2:
            with httpx.Client() as client:
                with client.stream("GET", url) as resp:
                    for chunk in resp.iter_bytes():
                        hash_md5.update(chunk)
                        file2.write(chunk)

        if checksum != hash_md5.hexdigest():
            raise KiaraProcessingException(
                f"Can't downloda file '{file_name}', invalid checksum: {checksum} != {hash_md5.hexdigest()}"
            )

        return target_file

    def process(self, inputs: ValueMap, outputs: ValueMap):

        import pyzenodo3

        include_metadata = inputs.get_value_data("include_metadata")

        doi = inputs.get_value_data("doi")
        zen = pyzenodo3.Zenodo()

        record = zen.find_record_by_doi(doi)

        path = FileBundle.create_tmp_dir()
        shutil.rmtree(path, ignore_errors=True)
        path.mkdir()
        for file_data in record.data["files"]:
            self.download_file(file_data, path)

        if include_metadata:
            metadata_filename = self.get_config_value("metadata_filename")
            metadata_file = path / metadata_filename
            metadata_file.write_bytes(orjson.dumps(record.data))

        bundle = FileBundle.import_folder(path.as_posix())
        outputs.set_value("file_bundle", bundle)
